
/* prob25.c						*/
/* program for LCAU25 option 't'			*/
/* calculate probabilities related to evolution		*/
/* Harold V. McIntosh, 10 August 1987			*/
/* 29 February 1988 - adapted from LCAU21 to LCAU22	*/
/* 11 March 1989 - adapted from LCAU23 to LCAU24	*/
/* 16 March 1989 - draw diagonal in B panel		*/
/* 18 March 1989 - Options n,N - totalistic statistics	*/
/* 15 September 1989 - entropy calculations		*/
/* 22 December 1991 - Bernstein curve fitting etc.	*/

/* references:							*/
/*								*/
/* W. John Wilbur, David J. Lipman and Shihab A. Shamma		*/
/* On the prediction of local patterns in cellular automata	*/
/* Physica 19D 397-410 (1986)					*/ 
/*								*/
/* Howard A. Gutowitz, Jonathan D. Victor and Bruce W. Knight	*/
/* Local structure theory for cellular automata			*/
/* Physica 28D 18-48 (1987)					*/

/*  Copyright (C) 1987  */
/*  Copyright (C) 1988  */
/*  Copyright (C) 1989  */
/*  Harold V. McIntosh  */

# include <math.h>
# include "lcau25.h"

extern char   ascrule[KK][KK][KK][KK][KK][KK][KK][KK][KK][KK][KK];
extern char   binrule[KK][KK][KK][KK][KK][KK][KK][KK][KK][KK][KK];
extern char   trule[TS];
extern char   arr1[AL];
extern int    co, dn;					/* connectivity, dimension */
extern int    pr, pl;					/* rule, line prob, mils   */
extern int    bi;					/* bernstein index	   */
extern double mf;					/* mouse increment	   */
extern double sm[TS];					/* selfconsistent nbrhd    */
extern double sn[TS];					/* selfconsistent nbrhd    */
extern double bx[BD];					/* bernstein factor        */
extern double bp[BD];					/* bernstein coefficients  */
extern double pv[10];					/* polynomial vector 	   */
extern double fpa, fpb, fpc;				/* fixed points		   */

# define GROW	   12	/* row for graph panels		*/
# define BROW	    3	/* row for bar charts		*/
# define EROW	    1	/* row for evolution synopsis	*/
# define AORG 	    0	/* x-origin  of contour plot    */
# define BORG 	  109	/* x-origin  of 2-block param   */
# define CORG 	  219	/* x-origin  of Bernstein plot  */

# include "prob2r.c"

/* ---------------------------------------------------- */
/* -            edit the probability screen           - */
/* ---------------------------------------------------- */

edtri() {char c; int i; double smq();

  videomode(COLGRAF);
  videopalette(YELREGR);
  prmenu();

  while (0<1) {
  videocursor(0,0,0); videoputc(':',2);
  if (istot()) {scrstr("  T: ",3); tscrrul(); scrstr("      LCAU25");}
    else if (isstot()) {scrstr("  S: ",3);
	sscrrul(0); scrstr(",",3);
	sscrrul(1); scrstr("   LCAU25");}
    else scrrul();
  videocursor(0,0,36); videoputc('?',2);
  c=kbdin();
  if (c == '\015') break;
  videocursor(0,0,38); videoputc(c,2);
  videocursor(0,0,36); videoputc(' ',0);
  switch (c) {
    case '+': videopalette(WHCYMAG); break;
    case '-': videopalette(YELREGR); break;
    case 'a': asfreq(3); break;
    case 'b': rprob(pl); break;
    case 'B': pl=numin(0); prmenu(); break;
    case 'c': traj(numin(0),8,3); break;
    case 'C': trajy(numin(0),0); break;
    case 'd': trajy(numin(0),2500); break;
    case 'D': trajy(numin(0),50000); break;
    case 'e': pevolve(); break;
    case 'E': plevolve(); break;
    case 'f': ranchay(ascrule,pr,DS); break;
    case 'F': pr=numin(0); prmenu(); break;
    case 'g': lifreq(50,1,2); break;
    case 'G': lifreq(200,1,1); break;
    case '.': lifreq(1,1,2); break;
    case 'h': egraf(CORG,0,100); break;
    case 'i': pidiff(100); break;
    case 'j': pjdiff(100); break;
    case 'k': pkdiff(100); break;
    case 'l': pldiff(100); break;
    case 'L': lesq(50,2); break;
    case 'H': lesq(3,2); break;
    case 'm': moncar(1,2); break;
    case 'M': moncar(50,1); break;
    case 'n': clep('m'); neisum(8*BROW,AL,25); break;
    case 'N': clep('m'); nbrsum(8*BROW,AL,10); break;
    case 'o': co=numin(0); prmenu(); break;
    case 'O': dn=numin(0); prmenu(); break;
    case 'p': ptseq(100); break;
    case 'P': hipro(100,0); break;
    case 's': lifreq(50,3,2); break;
    case 'S': lifreq(50,2,2); break;
    case 'X': bx[bi]=smq(); prmenu(); break;
    case 'I': bi=numin(0); prmenu(); break;
    case 'x': pdiff(100); break;
    case 'y': pdiff3(100); break;
    case 't': twoblockfreq(100); break;
    case '1': clep('m'); oneblfreq(8*BROW,300,48); break;
    case '2': clep('m'); twoblfreq(8*BROW,300,48); break;
    case '3': clep('m'); thrblfreq(8*BROW,300,48); break;
    case '4': clep('m'); foublfreq(8*BROW,300,48); break;
    case '?': trmenu(); break;
    case '!': prmenu(); break;
    case 'z': clep(kbdin()); break;
    case '/': videomode(COLGRAF); videopalette(YELREGR); break;
    case '*': bcprin(); break;
    case '#': parfix(); break;
    case '&': sctf(100,smq()); break;
    case '$': for (i=0; i<11; i++) sctf(100,0.0999*((double)i)); break;
    case '%': clep('m'); neidif(8*BROW); break;
    case 'K': neicon(); break;

    case MOD: gbmar(bi,1); if (bi>0) bi--; gbmar(bi,2); prmenu(); break;
    case MOU: gbmar(bi,1); if (bi<TS-1) bi++; gbmar(bi,2); prmenu(); break;
    case MOR: bx[bi]+=mf; prmenu(); break;
    case MOL: bx[bi]-=mf; prmenu(); break;
    case MOA: mf*=0.5; prmenu(); break;
    case MOB: mf*=2.0; prmenu(); break;
    case MOX: for (i=0; i<BD; i++) bx[i]=1.0; mf=0.05; prmenu(); break;

    default: break;
    } /* end switch */
  }   /* end while  */
  videopalette(WHCYMAG);
  videomode(T80X25);
}      /* end edtri  */


/* -------------------------------------- */
/* show the evolution of a planar section */
/* -------------------------------------- */

plevolve() {int i, j, i0, i1, i2, i3, i4;
  clep('a');
  gfram(AORG,0,100,3);
  asctobin();
  for (i=0; i<8; i++) for (j=0; j<8; j++) {
          i1=i3=i4=0;
	  for (i0=0; i0<8; i0++)
	  for (i2=0; i2<8; i2++)
    {
    videodot(199-4-i2-12*j,AORG+4+i0+12*i,arr1[ix5(i0,i1,i2,i3,i4)]);
    }
    onegen(AL);
    if (kbdst()) {kbdin(); return;}
  }
}




/* gather totalistic neighborhood sums for one generation */
/* mm = number of cells					  */
/* b[cc]  = sum counts					  */

onesum(b,mm,cc) int mm, cc, *b; {int i, j, k;
  for (i=0; i<cc; i++) b[i]=0;
  k=arr1[mm-5]+arr1[mm-4]+arr1[mm-3]+arr1[mm-2]+arr1[mm-1]
	+arr1[0]+arr1[1]+arr1[2]+arr1[3]+arr1[4]+arr1[5]; b[k]++;
  k=arr1[mm-4]+arr1[mm-3]+arr1[mm-2]+arr1[mm-1]+arr1[0]
	+arr1[1]+arr1[2]+arr1[3]+arr1[4]+arr1[5]+arr1[6]; b[k]++;
  k=arr1[mm-3]+arr1[mm-2]+arr1[mm-1]+arr1[0]+arr1[1]
	+arr1[2]+arr1[3]+arr1[4]+arr1[5]+arr1[6]+arr1[7]; b[k]++;
  k=arr1[mm-2]+arr1[mm-1]+arr1[0]+arr1[1]+arr1[2]
	+arr1[3]+arr1[4]+arr1[5]+arr1[6]+arr1[7]+arr1[8]; b[k]++;
  k=arr1[mm-1]+arr1[0]+arr1[1]+arr1[2]+arr1[3]
	+arr1[4]+arr1[5]+arr1[6]+arr1[7]+arr1[8]+arr1[9]; b[k]++;
  for (i=5; i<mm-6; i++) {k=0; for (j=i-5; j<i+6; j++) k+=arr1[j]; b[k]++;}
  k=arr1[mm-10]+arr1[mm-9]+arr1[mm-8]+arr1[mm-7]+arr1[mm-6]+arr1[mm-5]
	+arr1[mm-4]+arr1[mm-3]+arr1[mm-2]+arr1[mm-1]+arr1[0]; b[k]++;
  k=arr1[mm-9]+arr1[mm-8]+arr1[mm-7]+arr1[mm-6]+arr1[mm-5]+arr1[mm-4]
	+arr1[mm-3]+arr1[mm-2]+arr1[mm-1]+arr1[0]+arr1[1]; b[k]++;
  k=arr1[mm-8]+arr1[mm-7]+arr1[mm-6]+arr1[mm-5]+arr1[mm-4]+arr1[mm-3]
	+arr1[mm-2]+arr1[mm-1]+arr1[0]+arr1[1]+arr1[2]; b[k]++;
  k=arr1[mm-7]+arr1[mm-6]+arr1[mm-5]+arr1[mm-4]+arr1[mm-3]+arr1[mm-2]
	+arr1[mm-1]+arr1[0]+arr1[1]+arr1[2]+arr1[3]; b[k]++;
  k=arr1[mm-6]+arr1[mm-5]+arr1[mm-4]+arr1[mm-3]+arr1[mm-2]+arr1[mm-1]
	+arr1[0]+arr1[1]+arr1[2]+arr1[3]+arr1[4]; b[k]++;
}

/* Generate coefficients of 2nd generation Bernstein Polynomial */

berncoef(f) double *f; {
int i, i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia;

for (i=0; i<BD; i++) bp[i]=0.0;
for (i0=0; i0<KK; i0++)
for (i1=0; i1<KK; i1++)
for (i2=0; i2<KK; i2++)
for (i3=0; i3<KK; i3++)
for (i4=0; i4<KK; i4++)
for (i5=0; i5<KK; i5++)
for (i6=0; i6<KK; i6++)
for (i7=0; i7<KK; i7++)
for (i8=0; i8<KK; i8++)
for (i9=0; i9<KK; i9++)
for (ia=0; ia<KK; ia++)
  if (ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]=='1')
    bp[i0+i1+i2+i3+i4+i5+i6+i7+i8+i9+ia]+=1.0;
for (i=0; i<BD; i++) bp[i]*=f[i];
}

/* Generate coefficients of 3rd generation Bernstein Polynomial */

bernthrd() {
int i, i0, i1, i2, i3, i4, i5, i6, i7, i8, i9;
int ia, ib, ic, id, ie, jf, ig, ih, ii, ij, ik;
int j0, j1, j2, j3, j4, j5, j6, j7, j8, j9, ja;

asctobin();
for (i=0; i<BD; i++) bp[i]=0.0;
for (i0=0; i0<KK; i0++)
for (i1=0; i1<KK; i1++)
for (i2=0; i2<KK; i2++)
for (i3=0; i3<KK; i3++)
for (i4=0; i4<KK; i4++)
for (i5=0; i5<KK; i5++)
for (i6=0; i6<KK; i6++)
for (i7=0; i7<KK; i7++)
for (i8=0; i8<KK; i8++)
for (i9=0; i9<KK; i9++)
for (ia=0; ia<KK; ia++)
for (ib=0; ib<KK; ib++)
for (ic=0; ic<KK; ic++)
for (id=0; id<KK; id++)
for (ie=0; ie<KK; ie++)
for (jf=0; jf<KK; jf++)
for (ig=0; ig<KK; ig++)
for (ih=0; ih<KK; ih++)
for (ii=0; ii<KK; ii++)
for (ij=0; ij<KK; ij++)
for (ik=0; ik<KK; ik++) {
  j0=binrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia];
  j1=binrule[i1][i2][i3][i4][i5][i6][i7][i8][i9][ia][ib];
  j2=binrule[i2][i3][i4][i5][i6][i7][i8][i9][ia][ib][ic];
  j3=binrule[i3][i4][i5][i6][i7][i8][i9][ia][ib][ic][id];
  j4=binrule[i4][i5][i6][i7][i8][i9][ia][ib][ic][id][ie];
  j5=binrule[i5][i6][i7][i8][i9][ia][ib][ic][id][ie][jf];
  j6=binrule[i6][i7][i8][i9][ia][ib][ic][id][ie][jf][ig];
  j7=binrule[i7][i8][i9][ia][ib][ic][id][ie][jf][ig][ih];
  j8=binrule[i8][i9][ia][ib][ic][id][ie][jf][ig][ih][ii];
  j9=binrule[i9][ia][ib][ic][id][ie][jf][ig][ih][ii][ij];
  ja=binrule[ia][ib][ic][id][ie][jf][ig][ih][ii][ij][ik];
  i=i0+i1+i2+i3+i4+i5+i6+i7+i8+i9+ia+ib+ic+id+ie+jf+ig+ih+ii+ij+ik;
  if (ascrule[j0][j1][j2][j3][j4][j5][j6][j7][j8][j9][ja]=='1') bp[i]+=1.0;
  }
}

/* evaluate the nth generation Bernstein polynomial at point p */

double bern(p,n) double p; int n; {
int    i, j, d;
double q, s, pq[BD];
  d=10*n-9;
  q=1.0-p;
  for (i=0; i<=d; i++) pq[i]=1.0;
  for (i=1; i<=d; i++) {
    for (j=0; j<i; j++) pq[j]*=q;
    for (j=i; j<=d; j++) pq[j]*=p;
    }
  for (i=0, s=0.0; i<=d; i++) s+=bp[i]*pq[i];
  return(s);
}

/* display frequencies in ascrule in color l */

asfreq(l) int l; {
int    i, j, i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia;
int    stal5, stal4, stal3, stal2, stal1;
int    stac;
int    star1, star2, star3, star4, star5;	/* statistic counts */
int    stat[KK];
double staf[KK], t, pp;

videoscroll(BROW,0,BROW+8,40,0,0);
videocursor(0,BROW,0);
pp=1.0/((double)(KK*KK*KK*KK*KK*KK*KK*KK*KK*KK*KK));
stal5=stal4=stal3=stal2=stal1=0;
stac=0;
star1=star2=star3=star4=star5=0;
for (i=0; i<KK; i++) stat[i]=0;
for (i0=0; i0<KK; i0++)
for (i1=0; i1<KK; i1++)
for (i2=0; i2<KK; i2++)
for (i3=0; i3<KK; i3++)
for (i4=0; i4<KK; i4++)
for (i5=0; i5<KK; i5++)
for (i6=0; i6<KK; i6++)
for (i7=0; i7<KK; i7++)
for (i8=0; i8<KK; i8++)
for (i9=0; i9<KK; i9++)
for (ia=0; ia<KK; ia++) {
  j=ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]-'0';
  stat[j]++;
  if(j==i0) star5++;
  if(j==i1) star4++;
  if(j==i2) star3++;
  if(j==i3) star2++;
  if(j==i4) star1++;
  if(j==i5) stac++;
  if(j==i6) stal1++;
  if(j==i7) stal2++;
  if(j==i8) stal3++;
  if(j==i9) stal4++;
  if(j==ia) stal5++;
  }
for (i=0; i<KK; i++) printf("%2d    - %5.2f\n",i,((double)stat[i])*pp);
printf("\n");
printf("4left  - %5.2f      4right  - %5.2f\n",((double)stal4)*pp,((double)star4)*pp);
printf("3left  - %5.2f      3right  - %5.2f\n",((double)stal3)*pp,((double)star3)*pp);
printf("2left  - %5.2f      2right  - %5.2f\n",((double)stal2)*pp,((double)star2)*pp);
printf("1left  - %5.2f      1right  - %5.2f\n",((double)stal1)*pp,((double)star1)*pp);
printf("still  - %5.2f\n",((double)stac)*pp);
for (i=0; i<KK; i++) staf[i]=((double)stat[i])*pp;
j=199-(int)(100.0*staf[1]);
i=CORG+(int)(100.0*staf[1]);
if (j>=2) videodot(j-2,i,l);
if (i>=2) videodot(j,i-2,l);
videodot(j,i,1);
if (i<=318) videodot(j,i+2,l);
if (j<=198) videodot(j+2,i,l);
for (i=0, t=0.0; i<KK; i++) t+=staf[i]*log(staf[i]);
videocursor(0,BROW,20);
printf("lambda:   %5.3f",1.0-staf[0]);
videocursor(0,BROW+1,20);
printf("entropy:  %5.3f",-t/log((double)KK));
}

/* evaluate the parameters for two-block probabilities after one generation */

twoblock(z) double *z; {
int    i, i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia, ib;
int    j, j0, j1;
double w, x, y, p, q, pp;
double a[KK][KK], b[KK], c[KK][KK];

p=z[0]; q=1.0-p; pp=z[1];
a[1][1]=pp; a[1][0]=a[0][1]=p-pp; a[0][0]=1.0-2*p+pp;
b[1]=p; b[0]=q;
for (i=0; i<KK; i++) {for (j=0; j<KK; j++) c[i][j]=0.0;}
for (i0=0; i0<KK; i0++)
for (i1=0; i1<KK; i1++)
for (i2=0; i2<KK; i2++)
for (i3=0; i3<KK; i3++)
for (i4=0; i4<KK; i4++)
for (i5=0; i5<KK; i5++)
for (i6=0; i6<KK; i6++)
for (i7=0; i7<KK; i7++)
for (i8=0; i8<KK; i8++)
for (i9=0; i9<KK; i9++)
for (ia=0; ia<KK; ia++)
for (ib=0; ib<KK; ib++) {
  j0=ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]-'0';
  j1=ascrule[i1][i2][i3][i4][i5][i6][i7][i8][i9][ia][ib]-'0';
  x=a[i0][i1]*a[i1][i2]*a[i2][i3]*a[i3][i4]*a[i4][i5]*a[i5][i6]
	*a[i6][i7]*a[i7][i8]*a[i8][i9]*a[i9][ia]*a[ia][ib];
  y=b[i1]*b[i2]*b[i3]*b[i4]*b[i5]*b[i6]*b[i7]*b[i8]*b[i9]*b[ia];
  w=0.0; if (y>0.0001) w=x/y;
  c[j0][j1]+=w;
  }
z[0]=c[1][0]+c[1][1]; z[1]=c[1][1];
}

/* graph the 2-block parameter differences to find self-consistent values */
/* nxn points 								  */

twoblockfreq(n) int n; {
int i, j, l;
double op[2], np[2], s, t, u, v;

gfram(0,0,100,3);

s=1.0/((double)(n));
op[1]=s;
for (i=1; i<n; i++) {
  op[0]=s;
  for (j=1; j<n; j++) {
    np[0]=op[0]; np[1]=op[1]; twoblock(np);
    u=np[0]-op[0]; v=np[1]-op[1];
    t=u*u+v*v;
    if (t<0.0025) l=0; else {
    if (t<0.0675) {if((i+j)%2) l=0; else l=3;} else {
    if (t<0.3906) l=3; else {
    if (t<0.7500) {if((i+j)%2) l=3; else l=2;} else {
    if (t<1.5000) l=2; else {
    if (t<3.0000) {if((i+j)%2) l=2; else l=1;} else {
    if (t<6.0000) l=1; else {
                  if((i+j)%2) l=1; else l=0;
       }}}}}}}
    videodot(199-i,j,l);
    if (kbdst()) {kbdin(); return;}
    op[0]+=s;
    }  /* end for-j */
  op[1]+=s;
  }    /* end for-i */
}

/* evaluate the one-block probabilities after one generation */
onebl(x,a) double x[KK], a[KK]; {
int i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia, j0;

for (j0=0; j0<KK; j0++) x[j0]=0.0;

for (i0=0; i0<KK; i0++)
for (i1=0; i1<KK; i1++)
for (i2=0; i2<KK; i2++)
for (i3=0; i3<KK; i3++)
for (i4=0; i4<KK; i4++)
for (i5=0; i5<KK; i5++)
for (i6=0; i6<KK; i6++)
for (i7=0; i7<KK; i7++)
for (i8=0; i8<KK; i8++)
for (i9=0; i9<KK; i9++)
for (ia=0; ia<KK; ia++) {
  j0=ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]-'0';
  x[j0]+=a[i0]*a[i1]*a[i2]*a[i3]*a[i4]*a[i5]*a[i6]*a[i7]*a[i8]*a[i9]*a[ia];
  }
}

/* iterate the 1-block parameters to find self-consistent values */
/* graph the iterative steps in bar-chart form			 */
/* ll - initial line	*/
/* mm - length of line	*/
/* nn - number of lines	*/
oneblfreq(ll,mm,nn) int ll, mm, nn; {
int    ii, i, l, m, n;
double op[KK], np[KK];
double d, e, f, s;

m=0;
f=(double)mm;
s=1.0/((double)(KK));
n=(int)(f*s);
videodot(ll,m++,3);
for (i=0; i<KK; i++) {op[i]=s; for (l=0; l<n; l++) videodot(ll,m++,i);}

for (ii=1; ii<=nn; ii++) {
  e=0.0; m=0;
  onebl(np,op);
  videodot(ll+ii,m++,3);
  for (i=0; i<KK; i++) {
    n=(int)(f*np[i]); if (n>0) for (l=0; l<n; l++) videodot(ll+ii,m++,i);
    d=op[i]-np[i];
    e+=d*d;
    op[i]=np[i];
    }
  videodot(ll+ii,m++,3);
  if (op[0]<=0.001) break;
  if (op[1]<=0.001) break;
  if (e<=0.0000001) break;
  if (kbdst()) {kbdin(); break;}
  }
videocursor(0,BROW+7,0);
printf("(1-block) "); 
for (i=0; i<KK; i++) printf("%2d:%5.3f ",i,op[i]);
}

/* evaluate the two-block probabilities after one generation */
twobl(x,a) double x[KK][KK], a[KK][KK]; {
int    i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia, ib;
int    j0, j1;
double w, b[KK];

for (j0=0; j0<KK; j0++) {for (j1=0; j1<KK; j1++) x[j0][j1]=0.0;}

for (j0=0; j0<KK; j0++) {b[j0]=0.0; for (j1=0; j1<KK; j1++) b[j0]+=a[j0][j1];}

for (i0=0; i0<KK; i0++)
for (i1=0; i1<KK; i1++)
for (i2=0; i2<KK; i2++)
for (i3=0; i3<KK; i3++)
for (i4=0; i4<KK; i4++)
for (i5=0; i5<KK; i5++)
for (i6=0; i6<KK; i6++)
for (i7=0; i7<KK; i7++)
for (i8=0; i8<KK; i8++)
for (i9=0; i9<KK; i9++)
for (ia=0; ia<KK; ia++)
for (ib=0; ib<KK; ib++) {
  j0=ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]-'0';
  j1=ascrule[i1][i2][i3][i4][i5][i6][i7][i8][i9][ia][ib]-'0';
  w=a[i0][i1]*a[i1][i2]*a[i2][i3]*a[i3][i4]*a[i4][i5]
	*a[i5][i6]*a[i6][i7]*a[i7][i8]*a[i8][i9]*a[i9][ia]*a[ia][ib];
  if (w!=0.0) w/=b[i1]*b[i2]*b[i3]*b[i4]*b[i5]*b[i6]*b[i7]*b[i8]*b[i9]*b[ia];
  x[j0][j1]+=w;
  }
}

/* iterate the 2-block parameters to find self-consistent values */
/* graph the iterative steps in bar-chart form			 */
/* ll - initial line	*/
/* mm - length of line	*/
/* nn - number of lines	*/
twoblfreq(ll,mm,nn) int ll, mm, nn; {
int    ii, i, j, l, m, n;
double op[KK][KK], np[KK][KK];
double b[KK], d, e, f, s;

m=0;
f=(double)mm;
s=1.0/((double)(KK*KK));
n=(int)(f*s);
videodot(ll,m++,3);
for (i=0; i<KK; i++) {
for (j=0; j<KK; j++) {
  op[i][j]=s;
  for (l=0; l<n; l++) videodot(ll,m++,j);
  }}
videodot(ll,m++,3);

for (i=0; i<KK; i++) {b[i]=0.0; for (j=0; j<KK; j++) b[i]+=op[i][j];}
videodot(199-(int)(100.0*op[1][1]),(int)(100.0*b[1]),1);

for (ii=1; ii<=nn; ii++) {
  e=0.0; m=0;
  twobl(np,op);
  videodot(ll+ii,m++,3);
  for (i=0; i<KK; i++) {
  for (j=0; j<KK; j++) {
    n=(int)(f*np[i][j]);
    if (n>0) for (l=0; l<n; l++) videodot(ll+ii,m++,j);
    d=op[i][j]-np[i][j];
    e+=d<0.0?-d:d;
    op[i][j]=np[i][j];
    }}
  videodot(ll+ii,m++,3);
  for (i=0; i<KK; i++) {b[i]=0.0; for (j=0; j<KK; j++) b[i]+=op[i][j];}
  videodot(199-(int)(100.0*op[1][1]),(int)(100.0*b[1]),2);
  if (e<=0.0001) break;
  if (kbdst()) {kbdin(); break;}
  }
videocursor(0,BROW+7,0);
printf("(2-block) "); 
for (i=0; i<KK; i++) printf("%2d:%5.3f ",i,b[i]);
videocursor(0,BROW+8,0);
for (i=0; i<KK; i++) for (j=0; j<KK; j++) 
		printf("%1d%1d:%5.3f ",i,j,op[i][j]);
videodot(199-(int)(100.0*op[1][1]),(int)(100.0*b[1]),3);
videodot(199-(int)(100.0*op[1][1]),BORG+(int)(100.0*b[1]),1);
}

/* evaluate the three-block probabilities after one generation */
thrbl(x,a) double x[KK][KK][KK], a[KK][KK][KK]; {
int    i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia, ib, ic;
int    j0, j1, j2;
double w, b[KK][KK];

for (j0=0; j0<KK; j0++) {
for (j1=0; j1<KK; j1++) {
  for (j2=0; j2<KK; j2++) x[j0][j1][j2]=0.0;
  b[j0][j1]=0.0;
  for (j2=0; j2<KK; j2++) b[j0][j1]+=a[j0][j1][j2];  
  }}

for (i0=0; i0<KK; i0++)
for (i1=0; i1<KK; i1++)
for (i2=0; i2<KK; i2++)
for (i3=0; i3<KK; i3++)
for (i4=0; i4<KK; i4++)
for (i5=0; i5<KK; i5++)
for (i6=0; i6<KK; i6++)
for (i7=0; i7<KK; i7++)
for (i8=0; i8<KK; i8++)
for (i9=0; i9<KK; i9++)
for (ia=0; ia<KK; ia++)
for (ib=0; ib<KK; ib++)
for (ic=0; ic<KK; ic++) {
  j0=ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]-'0';
  j1=ascrule[i1][i2][i3][i4][i5][i6][i7][i8][i9][ia][ib]-'0';
  j2=ascrule[i2][i3][i4][i5][i6][i7][i8][i9][ia][ib][ic]-'0';
  w=a[i0][i1][i2]*a[i1][i2][i3]*a[i2][i3][i4]*a[i3][i4][i5]
	*a[i4][i5][i6]*a[i5][i6][i7]*a[i6][i7][i8]*a[i7][i8][i9]
	*a[i8][i9][ia]*a[i9][ia][ib]*a[ia][ib][ic];
  if (w!=0.0) w/=b[i1][i2]*b[i2][i3]*b[i3][i4]*b[i4][i5]
	*b[i5][i6]*b[i6][i7]*b[i7][i8]*b[i8][i9]*b[i9][ia]*b[ia][ib];
  x[j0][j1][j2]+=w;
  }
}

/* iterate the 3-block parameters to find self-consistent values */
/* ll - initial line	*/
/* mm - length of line	*/
/* nn - number of lines	*/
thrblfreq(ll,mm,nn) int ll, mm, nn; {
int    ii, i, j, k, l, m, n;
double op[KK][KK][KK], np[KK][KK][KK];
double b[KK], bb[KK][KK], d, e, f, s;

m=0;
f=(double)mm;
s=1.0/((double)(KK*KK*KK));
n=(int)(f*s);
videodot(ll,m++,3);
for (i=0; i<KK; i++) {
for (j=0; j<KK; j++) {
for (k=0; k<KK; k++) {
  op[i][j][k]=s;
  for (l=0; l<n; l++) videodot(ll,m++,k);
  }}}
videodot(ll,m++,3);

for (ii=1; ii<=nn; ii++) {
  e=0.0; m=0;
  thrbl(np,op);
  videodot(ll+ii,m++,3);
  for (i=0; i<KK; i++) {
  for (j=0; j<KK; j++) {
  for (k=0; k<KK; k++) {
    n=(int)(f*np[i][j][k]);
    if (n>0) for (l=0; l<n; l++) videodot(ll+ii,m++,k);
    d=op[i][j][k]-np[i][j][k];
    e+=d<0.0?-d:d;
    op[i][j][k]=np[i][j][k];
    }}}
  videodot(ll+ii,m++,3);
  if (e<=0.0001) break;
  if (kbdst()) {kbdin(); break;}
  }

for (i=0; i<KK; i++) {
  b[i]=0.0;
  for (j=0; j<KK; j++) {
  for (k=0; k<KK; k++) {
    b[i]+=op[i][j][k];
    }} }
videocursor(0,BROW+7,0);
printf("(3-block) "); 
for (i=0; i<KK; i++) printf("%2d:%5.3f ",i,b[i]);

for (i=0; i<KK; i++) {
for (j=0; j<KK; j++) {
  bb[i][j]=0.0;
  for (k=0; k<KK; k++) {
    bb[i][j]+=op[i][j][k];
    }} }

videocursor(0,BROW+8,0);
for (i=0; i<KK; i++) for (j=0; j<KK; j++) 
  printf("%1d%1d:%5.3f ",i,j,bb[i][j]);
videodot(199-(int)(100.0*bb[1][1]),BORG+(int)(100.0*b[1]),2);
}

/* evaluate the four-block probabilities after one generation */
foubl(x,a) double x[KK][KK][KK][KK], a[KK][KK][KK][KK]; {
int    i0, i1, i2, i3, i4, i5, i6, i7, i8, i9, ia, ib, ic, id;
int    j0, j1, j2, j3;
double w, b[KK][KK][KK];

for (j0=0; j0<KK; j0++) {
for (j1=0; j1<KK; j1++) {
for (j2=0; j2<KK; j2++) {
  for (j3=0; j3<KK; j3++) x[j0][j1][j2][j3]=0.0;
  b[j0][j1][j2]=0.0;
  for (j3=0; j3<KK; j3++) b[j0][j1][j2]+=a[j0][j1][j2][j3];  
  }}}

for (i0=0; i0<KK; i0++)
for (i1=0; i1<KK; i1++)
for (i2=0; i2<KK; i2++)
for (i3=0; i3<KK; i3++)
for (i4=0; i4<KK; i4++)
for (i5=0; i5<KK; i5++)
for (i6=0; i6<KK; i6++)
for (i7=0; i7<KK; i7++)
for (i8=0; i8<KK; i8++)
for (i9=0; i9<KK; i9++)
for (ia=0; ia<KK; ia++)
for (ib=0; ib<KK; ib++)
for (ic=0; ic<KK; ic++)
for (id=0; id<KK; id++) {
  j0=ascrule[i0][i1][i2][i3][i4][i5][i6][i7][i8][i9][ia]-'0';
  j1=ascrule[i1][i2][i3][i4][i5][i6][i7][i8][i9][ia][ib]-'0';
  j2=ascrule[i2][i3][i4][i5][i6][i7][i8][i9][ia][ib][ic]-'0';
  j3=ascrule[i3][i4][i5][i6][i7][i8][i9][ia][ib][ic][id]-'0';
  w=a[i0][i1][i2][i3];
  w*=a[i1][i2][i3][i4];
  w*=a[i2][i3][i4][i5];
  w*=a[i3][i4][i5][i6];
  w*=a[i4][i5][i6][i7];
  w*=a[i5][i6][i7][i8];
  w*=a[i6][i7][i8][i9];
  w*=a[i7][i8][i9][ia];
  w*=a[i8][i9][ia][ib];
  w*=a[i9][ia][ib][ic];
  w*=a[ia][ib][ic][id];
  if (w!=0.0) w/=b[i1][i2][i3]*b[i2][i3][i4]*b[i3][i4][i5]
	*b[i4][i5][i6]*b[i5][i6][i7]*b[i6][i7][i8]*b[i7][i8][i9]
	*b[i8][i9][ia]*b[i9][ia][ib]*b[ia][ib][ic];
  x[j0][j1][j2][j3]+=w;
  }
}

/* iterate the 4-block parameters to find self-consistent values */
/* ll - initial line	*/
/* mm - length of line	*/
/* nn - number of lines	*/
foublfreq(ll,mm,nn) int ll, mm, nn; {
int    ii, i0, i1, i2, i3, l, m, n;
double op[KK][KK][KK][KK], np[KK][KK][KK][KK];
double b[KK], bb[KK][KK], d, e, f, s;

m=0;
f=(double)mm;
s=1.0/((double)(KK*KK*KK*KK));
n=(int)(f*s);
videodot(ll,m++,3);
for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
for (i2=0; i2<KK; i2++) {
for (i3=0; i3<KK; i3++) {
  op[i0][i1][i2][i3]=s;
  for (l=0; l<n; l++) videodot(ll,m++,i3);
  }}}}
videodot(ll,m++,3);

for (ii=1; ii<=nn; ii++) {
  e=0.0; m=0;
  foubl(np,op);
  videodot(ll+ii,m++,3);
  for (i0=0; i0<KK; i0++) {
  for (i1=0; i1<KK; i1++) {
  for (i2=0; i2<KK; i2++) {
  for (i3=0; i3<KK; i3++) {
    n=(int)(f*np[i0][i1][i2][i3]);
    if (n>0) for (l=0; l<n; l++) videodot(ll+ii,m++,i3);
    d=op[i0][i1][i2][i3]-np[i0][i1][i2][i3];
    e+=d<0.0?-d:d;
    op[i0][i1][i2][i3]=np[i0][i1][i2][i3];
    }}}}
  videodot(ll+ii,m++,3);
  if (e<=0.0001) break;
  if (kbdst()) {kbdin(); break;}
  }

for (i0=0; i0<KK; i0++) {
  b[i0]=0.0;
  for (i1=0; i1<KK; i1++) {
  for (i2=0; i2<KK; i2++) {
  for (i3=0; i3<KK; i3++) {
    b[i0]+=op[i0][i1][i2][i3];
    }}} }

videocursor(0,BROW+7,0);
printf("(4-block) "); 
for (i0=0; i0<KK; i0++) printf("%2d:%5.3f ",i0,b[i0]);

for (i0=0; i0<KK; i0++) {
for (i1=0; i1<KK; i1++) {
  bb[i0][i1]=0.0;
  for (i2=0; i2<KK; i2++) {
  for (i3=0; i3<KK; i3++) {
    bb[i0][i1]+=op[i0][i1][i2][i3];
    }} }}

videocursor(0,BROW+8,0);
for (i0=0; i0<KK; i0++) for (i1=0; i1<KK; i1++) 
  printf("%1d%1d:%5.3f ",i0,i1,bb[i0][i1]);
videodot(199-(int)(100.0*bb[1][1]),BORG+(int)(100.0*b[1]),2);
}

/* end */

